/*
 *	$Header: /home/glenn/src/labpc/RCS/adc.c,v 0.2 1995/10/24 16:32:20 glenn Exp glenn $
 *
 *	$Log: adc.c,v $
 *	Revision 0.2  1995/10/24 16:32:20  glenn
 *	Revision 0.2
 *
 */

#include <linux/config.h>

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif MODULE

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/config.h>
#include <linux/string.h>
#include <linux/malloc.h>
#include <linux/mm.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/dma.h>

#include "adc.h"
#include "counter.h"
#include "labpc.h"

#define MAX_DMA_SIZE	(64 * 1024)
#define MAX_COUNT	(64 * 1024)

#ifndef ADCBUFSIZE
#define ADCBUFSIZE	1024
#endif

static struct semaphore mutex = MUTEX;

int adcbufsize = ADCBUFSIZE;

/* Clear the ADC - write to clear register and read 2 values from FIFO */
#define ClearADC( adc )			\
	{ outb_p( 0x00, (adc)->clear );	\
	inb_p( (adc)->fifo );		\
	inb_p( (adc)->fifo );		\
	}

/* Start an ADC conversion by writing to the ADC start register */
#define StartADCConversion( adc )	\
	outb_p( 0x00, (adc)->start )

/* Read a 16-bit ADC conversion value from the ADC FIFO */
#define ReadADCFIFO( adc )		\
	(inb_p( (adc)->fifo ) | (inb_p( (adc)->fifo ) << 8))

/* Accessing Board command and status registers */
#define ADCstatus( board )		\
	(readport( &((board)->regs.stat) ))

/* Enable ADC DMA interrupt generation */
#define EnableDMAInterrupts( adc )	\
	setbits( &((board)->regs.cmd3), DMAEN | TCINTEN | ERRINTEN )
/* Clear an ADC DMA Terminal Count interrupt */
#define ClearDMAInterrupt( adc )	\
	outb_p( 0x00, (adc)->dmaintclr )

/* Enable interrupts on ADC FIFO */
#define EnableADCInterrupts( adc )	\
	setbits( &((board)->regs.cmd3), FIFOINTEN | ERRINTEN )

/* Disable all ADC interrupt generation */
#define DisableADCInterrupts( adc )		\
	clearbits( &((board)->regs.cmd3),	\
		  DMAEN | TCINTEN | ERRINTEN | FIFOINTEN )

#define STATIC

STATIC int allocate_adc_buffers( ADC_t *adc )
{
	if (adc->bufsize == 0) {
		adc->frontbuf = NULL;
		adc->backbuf = NULL;
		return 0;
	}

	/* Allocate buffers for DMA transfers */
	if (adc->bufsize * sizeof( short ) > MAX_DMA_SIZE) {
		printk( "LabPC+: Requested DMA buffer size, %d, too large\n",
		       adc->bufsize );
		adc->bufsize = MAX_DMA_SIZE / sizeof( short );
		printk( "LabPC+: Using DMA buffer size, %d\n",
		       adc->bufsize );
	}
	adc->frontbuf = kmalloc( adc->bufsize * sizeof( short ),
				 GFP_KERNEL | GFP_DMA );
	if (adc->frontbuf == NULL) {
		printk( "ADC: Unable to allocate DMA buffer\n" );
		adc->bufsize = 0;
		return -(ENOMEM);
	}
	adc->backbuf = kmalloc( adc->bufsize * sizeof( short ),
			        GFP_KERNEL | GFP_DMA );
	if (adc->backbuf == NULL) {
		kfree( adc->frontbuf );
		adc->bufsize = 0;
		printk( "ADC: Unable to allocate DMA buffer\n" );
		return -(ENOMEM);
	}

	return 0;
}

STATIC int read_adc( struct inode *inode, struct file *file,
		     char *buffer, int count )
{
	ADC_t	*adc;
	int 	retval = 0;
	int	status;
	int	dynamicbuffer;

	dprintregs();
	
	/* Which ADC? */
	if ((adc = GETADC( MINOR(inode->i_rdev) )) == NULL)
		return -(EINVAL);

	count /= sizeof( short );

	if (count < 1)
		return -(EINVAL);

	/* Lockout other processes from accessing the ADC */
	down( &(adc->sem) );

	adc->waitfortrig = 0;
	adc->overflow = adc->overrun = 0;

	dynamicbuffer = 0;
	if (adc->bufsize == 0) {
		/* Dynamically allocate a buffer for DMA transfers */
		dynamicbuffer = 1;
		adc->frontbuf = kmalloc( count * sizeof( short ),
					 GFP_KERNEL | GFP_DMA );
		if (adc->frontbuf == NULL) {
			printk( "ADC: Unable to allocate DMA buffer\n" );
			up( &(adc->sem) );
			return -(ENOMEM);
		}
		adc->bufsize = count;
	}

	/* Disable all ADC interrupts (DMA, error, fifo) */
	DisableADCInterrupts( adc );

	clearbits( &(board->regs.cmd2), SWTRIG | HWTRIG | PRETRIG );

	if (count > 1) {
		/* This follows the recommended sequence in
		 * Chapter 5 of the Lab-PC+ manual
		 * ("Programming Multiple A/D Conversions with
		 * Channel Scanning"). I clear the bit twice
		 * just to be sure - this may not be
		 * necessary.  */
		Counter_t *counter;

		if (adc->scan) {
			clearbits( &(board->regs.cmd1), SCANEN );
			clearbits( &(board->regs.cmd1), SCANEN );
			setbits( &(board->regs.cmd1), SCANEN );
		}

		counter = GETCOUNTER( DAQ_COUNTERA1 );
		SetHigh8253( counter );
		if (adc->counter_val) {
			/* Set the interval counter for the ADC */
			counter = GETCOUNTER( DAQ_COUNTERA0 );
			SetMode8253( counter, MODE_RATE_GENERATOR );
			Write8253( counter, adc->counter_val );

			if (adc->timebase != 2) {
				counter = GETCOUNTER( DAQ_COUNTERB0 );
				SetMode8253( counter, MODE_SQUARE_WAVE );
				Write8253( counter, adc->timebase );
			}
		} else {
			/* Use EXTCONV* for external ADC trigger */
			clearbits( &(board->regs.cmd4), ECKDRV );
			counter = GETCOUNTER( DAQ_COUNTERA0 );
			SetHigh8253( counter );
		}

		ClearADC( adc );
		
		counter = GETCOUNTER( DAQ_COUNTERA1 );
		if (count <= MAX_COUNT) {
			/* Set the sample counter for the ADC */
			SetMode8253( counter, MODE_TERMINAL_COUNT );
			Write8253( counter, count - 2 );
		} else {
			/* Program in Freerun mode - set A1 output low */
			SetLow8253( counter );
		}
	}

	cli(); /* interrupts re-enabled by interruptible_sleep_on() */
	adc->residue = count;
	adc->count = (adc->residue > adc->bufsize) ? adc->bufsize : adc->residue;
	adc->indx = adc->incr = 0;
	if (adc->dmaen) {
		disable_dma( board->dma );

		/* Enable LabPC card for DMA and terminal count interrupts */
		EnableDMAInterrupts( adc );

		clear_dma_ff( board->dma );
		set_dma_addr( board->dma, (unsigned long)adc->frontbuf );
		set_dma_count( board->dma, adc->count * sizeof( short ) );
		set_dma_mode( board->dma, DMA_MODE_READ );
		enable_dma( board->dma );
	} else {
		EnableADCInterrupts( adc );
	}

	retval = 0;
	switch (adc->triggermode) {
	case ADC_AUTOTRIGGER:
		/* Trigger the data acquisition sequence */
		if (count > 1) {
			setbits( &(board->regs.cmd2), SWTRIG );
		} else {
			StartADCConversion( adc );
		}
		break;
	case ADC_SOFTTRIGGER:
		/* Just start the read and wait for software ADC trigger */
		adc->waitfortrig = 1;
		break;
	case ADC_HARDTRIGGER:
		/* Just start the read and wait for hardware ADC trigger */
		break;
	case ADC_PRETRIGGER:
		sti();
		dprintk(( "ADC: trigger type not implemented\n" ));
		retval = -(EINVAL);
		break;
	default:
		sti();
		retval = -(EIO);
		break;
	}

	while (adc->residue && retval == 0) {
		interruptible_sleep_on( &(adc->wait) );

		status = ADCstatus( board );
		if (current->signal & ~current->blocked) {
			ClearADC( adc );
			dprintk(( "ADC: interrupted\n" ));
			retval = -(ERESTARTSYS);
			break;
		}
		if (adc->overrun) {
			retval = -(EIO);
			dprintregs();
			printk( "ADC: overrun detected\n" );
			break;
		} else if (adc->overflow) {
			retval = -(EIO);
			dprintregs();
			printk( "ADC: overflow detected\n" );
			break;
		} else if (adc->indx) {
			/* A buffer has filled - copy it to user space */
			memcpy_tofs( buffer, adc->backbuf, adc->indx * sizeof( short ) );
			buffer += adc->indx * sizeof( short );
			adc->indx = 0;
		} else {
			dprintregs();
			printk( "LabPC+: DMA buffer overflow\n" );
			retval = -(EIO);
			break;
		}
	}
	adc->waitfortrig = 0;

	/* Disable DMA transfer and terminal count interrupts */
	DisableADCInterrupts( adc );

	/* Clear the trigger condition */
	clearbits( &(board->regs.cmd2), SWTRIG );

	ClearADC( adc );
	if (adc->dmaen)
		disable_dma( board->dma );

	if (dynamicbuffer) {
		kfree( adc->frontbuf );
		adc->frontbuf = NULL;
		adc->bufsize = 0;
	}

	/* Release the exclusion semaphore */
	up( &(adc->sem) );
	
	if (retval == 0) {
		dprintk(( "ADC: %d DMA words transferred\n", count ));
			 
		retval = count * sizeof( short );
	}

	return (retval);
}

STATIC int select_adc( struct inode *inode, struct file *file,
		       int sel_type, select_table *wait )
{
	ADC_t *adc = BOARD( MAJOR( inode->i_rdev ) )->adc;

	if (sel_type != SEL_IN)
		return 1;

	cli();
	if (ADCstatus( board ) & DAVAIL)
		return 1;

	select_wait( &(adc->wait), wait );
	sti();

	return 0;
}

STATIC int init_adc( ADC_t *adc )
{
	dprintregs();
	DisableADCInterrupts( adc );
	ClearADC( adc );

	/* Initialise the ADC registers */

	/* ADC channel 0: ADC output -2048 to 2047 : Gain = 0 : Disable Scan */
	clearbits( &(board->regs.cmd1), (MA0 | MA1 | MA2 | TWOSCMP |
				     GAIN0 | GAIN1 | GAIN2 | SCANEN) );
	setbits( &(board->regs.cmd1), TWOSCMP );
	/* Disable pretrigger: Disable Hardware trigger :
	 * Disable Software trigger: Time Base = 0 */
	clearbits( &(board->regs.cmd2), (PRETRIG | HWTRIG | SWTRIG) );
	/* Enable ADC interrupts on error and FIFO data ready - Disable DMA  */
	DisableADCInterrupts( adc );
	/* Disable Interval scanning: Disable clock for interval scanning:
	 * Direction of EXTCONV : Single ended ADC input */
	clearbits( &(board->regs.cmd4), (INTSCAN | EOIRCV | ECKDRV | SE_D) );

	/* Set the data ports for the ADC */
	adc->fifo	= DAQ_FIFO;
	adc->clear	= DAQ_ADCCLR;
	adc->start	= DAQ_STRTCVT;
	adc->dmaintclr	= DAQ_DMATCCLR;

	/* Initialise options and state variables */
	adc->scan	= ADC_SCAN_OFF;
	adc->gain	= ADC_GAIN_1;
	adc->polarity	= ADC_BIPOLAR;
	adc->channel	= 0;
	adc->timebase	= 2;
	adc->triggermode= ADC_AUTOTRIGGER;
	adc->dmaen	= 1;
	adc->inputmode	= 0;
	adc->eckdrv	= 0;
	adc->eoircv	= 0;
	adc->intscan	= 0;
	adc->counter_val= 1000;
	adc->waitfortrig= 0;

	/* Access control variables */
	adc->sem   	= mutex;
	adc->wait  	= NULL;
	adc->overflow 	= 0;
	adc->overrun	= 0;

	dprintregs();

	return 0;
}

STATIC int ioctl_adc( struct inode *inode, struct file *file,
		      unsigned int cmd, unsigned long arg )
{
	ADC_t 	*adc;
	Counter_t *counter;
	int	stat = 0;
	
	if ((stat = labpc_ioctl_check( cmd, arg )) != 0)
		return stat;
		
	/* Which ADC? */
	if ((adc = GETADC( MINOR(inode->i_rdev) )) == NULL)
		return -(EINVAL);

	/* Set the counter for the DAC input */
	counter = GETCOUNTER( DAQ_COUNTERA0 );

	switch (cmd) {
	case ADC_SET_SCAN:
		/* Enable/disable ADC input channel scanning */
		down( &(adc->sem) );
		if (arg == ADC_SCAN_ON) {
			setbits( &(board->regs.cmd1), SCANEN );
			adc->scan = arg;
		} else if (arg == ADC_SCAN_OFF) {
			clearbits( &(board->regs.cmd1), SCANEN );
			adc->scan = arg;
		} else {
			stat = -(EINVAL);
		}
		up( &(adc->sem) );
		break;
	case ADC_GET_SCAN:
		/* Get the DAC clock counter value */
		put_fs_long( adc->scan, (unsigned long *)arg );
		break;
	case ADC_SET_GAIN:
		/* Set the ADC gain */
		down( &(adc->sem) );
		if (arg > ADC_GAIN_MAX) {
			stat = -(EINVAL);
		} else {
			writebits( &(board->regs.cmd1),
				  (arg << 4) & (GAIN0 | GAIN1 | GAIN2),
				  (~(arg << 4)) & (GAIN0 | GAIN1 | GAIN2) );
			adc->gain = arg;
		}
		up( &(adc->sem) );
		break;
	case ADC_GET_GAIN:
		/* Get the DAC clock counter value */
		put_fs_long( adc->gain, (unsigned long *)arg );
		break;
	case ADC_SET_POLARITY:
		down( &(adc->sem) );
		if (arg == ADC_BIPOLAR) {
			/* Signal is bipolar - read data in 2's complement */
			setbits( &(board->regs.cmd1), TWOSCMP );
			adc->polarity = arg;
		} else if (arg == ADC_UNIPOLAR) {
			/* Signal is unipolar - read data in unsigned binary */
			clearbits( &(board->regs.cmd1), TWOSCMP );
			adc->polarity = arg;
		} else {
			stat = -(EINVAL);
		}
		up( &(adc->sem) );
		break;
	case ADC_GET_POLARITY:
		/* Get the DAC clock counter value */
		put_fs_long( adc->polarity, (unsigned long *)arg );
		break;
	case ADC_SET_CHANNEL:
		/* ADC input channel no. OR max channel for scanning */
		down( &(adc->sem) );
		if (arg >= ADC_NUM_CHANNELS) {
			stat = -(EINVAL);
		} else {
			writebits( &(board->regs.cmd1),
				  arg & (MA0 | MA1 | MA2),
				  (~arg) & (MA0 | MA1 | MA2) );
			adc->channel = arg;
		}
		up( &(adc->sem) );
		break;
	case ADC_GET_CHANNEL:
		put_fs_long( adc->channel, (unsigned long *)arg );
		break;
	case ADC_SET_TIMEBASE:
		down( &(adc->sem) );
		if (arg == 2) {
			adc->timebase = arg;
			clearbits( &(board->regs.cmd2), TBSEL );
		} else if (arg > 2) {
			adc->timebase = arg;
			setbits( &(board->regs.cmd2), TBSEL );
		} else {
			stat = -(EINVAL);
		}
		up( &(adc->sem) );
		break;
	case ADC_GET_TIMEBASE:
		put_fs_long( adc->timebase, (unsigned long *)arg );
		break;
	case ADC_SET_TRIGGER_MODE:
		/* Set the triggering mode for data acquisition */
		down( &(adc->sem) );
		switch (arg) {
		case ADC_SOFTTRIGGER:
			adc->triggermode = arg;
			clearbits( &(board->regs.cmd2),
				   (SWTRIG | HWTRIG | PRETRIG) );
			break;
		case ADC_HARDTRIGGER:
			adc->triggermode = arg;
			writebits( &(board->regs.cmd2),
				   HWTRIG, (SWTRIG | PRETRIG) );
			break;
		case ADC_PRETRIGGER:
			adc->triggermode = arg;
			writebits( &(board->regs.cmd2),
				   PRETRIG, (SWTRIG | HWTRIG) );
			break;
		case ADC_AUTOTRIGGER:
			adc->triggermode = arg;
			clearbits( &(board->regs.cmd2),
				   (PRETRIG | SWTRIG | HWTRIG) );
			break;
		default:
			stat = -(EINVAL);
		}
		up( &(adc->sem) );
		break;
	case ADC_GET_TRIGGER_MODE:
		/* Get the ADC trigger mode */
		put_fs_long( adc->triggermode, (unsigned long *)arg );
		break;
	case ADC_SET_CLOCK:
		adc->counter_val = arg;
		if (arg > 0) {
			stat = Write8253( counter, arg );
		} else {
			stat = SetHigh8253( counter );
		}
		break;
	case ADC_GET_CLOCK:
		/* Get the ADC clock counter value */
		put_fs_long( adc->counter_val, (unsigned long *)arg );
		break;
	case ADC_DMAEN:
		/* Enable DMA for data t/fers */
		down( &(adc->sem) );
		adc->dmaen = (arg != 0);
		up( &(adc->sem) );
		break;
	case ADC_SE_D:
		/* ADC inputs are single ended/differential */
		down( &(adc->sem) );
		if ((adc->inputmode = (arg != 0))) {
			setbits( &(board->regs.cmd4), SE_D );
		} else {
			clearbits( &(board->regs.cmd4), SE_D );
		}
		up( &(adc->sem) );
		break;
	case ADC_ECKDRV:
		/* Mode for the EXTCONV line on the I/O connector */
		down( &(adc->sem) );
		if ((adc->eckdrv = (arg != 0))) {
			setbits( &(board->regs.cmd4), ECKDRV );
		} else {
			clearbits( &(board->regs.cmd4), ECKDRV );
		}
		up( &(adc->sem) );
		break;
	case ADC_EOIRCV:
		/* Clock source for interval scanning */
		down( &(adc->sem) );
		if ((adc->eoircv = (arg != 0))) {
			setbits( &(board->regs.cmd4), EOIRCV );
		} else {
			clearbits( &(board->regs.cmd4), EOIRCV );
		}
		up( &(adc->sem) );
		break;
	case ADC_INTSCAN:
		/* Enable Interval Scanning */
		down( &(adc->sem) );
		adc->intscan = (arg != 0);
		up( &(adc->sem) );
		break;
	case ADC_INIT:
		down( &(adc->sem) );
		init_adc( adc );
		up( &(adc->sem) );
		break;
	case ADC_CLEAR:
		/* Clear the ADC fifo buffer */
		down( &(adc->sem) );
		ClearADC( adc );
		up( &(adc->sem) );
		break;
	case ADC_TRIGGER:
		/* Trigger an ADC conversion sequence */
		if (adc->triggermode == ADC_SOFTTRIGGER && adc->waitfortrig) {
			if (adc->count > 1) {
				setbits( &(board->regs.cmd2), SWTRIG );
			} else {
				StartADCConversion( adc );
			}
		} else {
			stat = -(EINVAL);
		}
		break;
	case ADC_SETBUFSIZE:
		if (arg * sizeof( short ) > MAX_DMA_SIZE) {
			return -(EINVAL);
		} else {
			down( &(adc->sem) );
			if (adc->bufsize) {
				kfree( adc->frontbuf );
				kfree( adc->backbuf );
			}	
			adc->bufsize = arg;
			stat = allocate_adc_buffers( adc );
			up( &(adc->sem) );
		}
		break;
	case ADC_GETBUFSIZE:
		put_fs_long( adc->bufsize, (unsigned long *)arg );
		break;
	case ADC_STATUS:
		/* Read the ADC status register - returning byte read */
		put_fs_long( (unsigned long)(ADCstatus( board ) & 0x7f),
			     (unsigned long *)arg );
		break;
	default:
		return -(ENOSYS);
		break;
	}

	return stat;
}

STATIC int open_adc( struct inode *inode, struct file *file )
{
	ADC_t 	*adc;
	
	dprintk(( "LabPC+: open_adc()\n" ));
		
	/* Which ADC? */
	if ((adc = GETADC( MINOR(inode->i_rdev) )) == NULL)
		return -(EINVAL);

	adc->open++;

	return 0;
}

STATIC void release_adc(struct inode *inode, struct file *file)
{
	ADC_t 		*adc;
	int		minor;

	dprintk(( "LabPC+: release_adc()\n" ));

	minor = MINOR(inode->i_rdev);
	if ((adc = GETADC( minor )) == NULL)
		return;

	adc->open--;

	if (adc->open == 0) {
		/* Last close - shutdown the ADC if in running mode */
		if (ADCstatus( board ) & (DAVAIL | ADC_OVERRUN | ADC_OVERFLOW))
			ClearADC( adc );
	}

	labpc_close( inode, file );
}

void adc_interrupt( ADC_t *adc, int status )
{
	short	*buf;

	if (status & DMATC) {
		if (!adc->dmaen) {
			printk( "ADC DMATC bit set when DMA transfer mode not enabled!\n" );
			ClearDMAInterrupt( adc );
			return;
		}
		adc->residue -= adc->count;

		/* Check the back buffer has been emptied */ 
		if (adc->indx) {
			printk( "LabPC+: ADC DMA buffer overflow\n" );
			adc->indx = 0;
			ClearDMAInterrupt( adc );
			wake_up_interruptible( &(adc->wait) );
			return;
		}
		buf = adc->frontbuf;
		adc->frontbuf = adc->backbuf;
		adc->backbuf = buf;
		adc->indx = adc->count;

		if (adc->residue) {
			/* setup next DMA transfer */
			adc->count = (adc->residue > adc->bufsize) ? adc->bufsize : adc->residue;

			set_dma_addr( board->dma, (unsigned long)adc->frontbuf );
			set_dma_count( board->dma, adc->count * sizeof( short ) );
			enable_dma( board->dma );
		}
		ClearDMAInterrupt( adc );

		wake_up_interruptible( &(adc->wait) );
	} else if (status & DAVAIL) {
		if (adc->dmaen) {
			printk( "ADC DAVAIL bit set when using DMA transfer mode!\n" );
			return;
		}
		((short *)adc->frontbuf)[adc->incr++] = ReadADCFIFO( adc );
		if (adc->incr >= adc->count) {
			adc->residue -= adc->count;

			/* Check the back buffer has been emptied */ 
			if (adc->indx) {
				printk( "LabPC+: ADC buffer overflow\n" );
				adc->indx = 0;
				wake_up_interruptible( &(adc->wait) );
				return;
			}
			buf = adc->frontbuf;
			adc->frontbuf = adc->backbuf;
			adc->backbuf = buf;
			adc->indx = adc->count;
			adc->incr = 0;
			if (adc->residue) {
				/* setup next DMA transfer */
				adc->count = (adc->residue > adc->bufsize) ? adc->bufsize : adc->residue;
			}
			wake_up_interruptible( &(adc->wait) );
		}
	} else if (status & (ADC_OVERRUN | ADC_OVERFLOW)) {
		dprintregs();
		DisableADCInterrupts( adc );
		ClearADC( adc );
		if (status & ADC_OVERRUN)
			adc->overrun = 1;
		if (status & ADC_OVERFLOW)
			adc->overflow = 1;
		wake_up_interruptible( &(adc->wait) );
	} else {
		dprintk(( "Lab-PC+: ADC interrupt for no good reason\n" ));
	}

	return;
}

struct file_operations adc_fops = {
	NULL,			/* seek */
	read_adc,
	NULL,			/* write */
	NULL,			/* readdir */
	select_adc,
	ioctl_adc,		/* ioctl */
	NULL,			/* mmap */
	open_adc,
	release_adc,
};

/* Functions to load and unload the ADC structure */

ADC_t *get_adc( int minor )
{
	ADC_t	*adc;
	
	adc = (ADC_t *)kmalloc( sizeof( ADC_t ), GFP_KERNEL );

	adc->open = 0;
	init_adc( adc );

	adc->bufsize = adcbufsize;
	if (allocate_adc_buffers( adc ) < 0) {
		kfree( adc );
		return NULL;
	}
	adc->indx = adc->count = 0;

	return adc;
}

void free_adc( ADC_t *adc )
{
	if (adc != NULL) {
		if (adc->frontbuf != NULL) {
			kfree( adc->frontbuf );
			adc->frontbuf = NULL;
		}

		if (adc->backbuf != NULL) {
			kfree( adc->backbuf );
			adc->backbuf = NULL;
		}
		init_adc( adc );

		kfree( adc );
	}
}

