
/*
 *	$Header: /home/glenn/src/labpc/RCS/dac.c,v 0.2 1995/10/24 16:32:20 glenn Exp glenn $
 *	$Log: dac.c,v $
 *	Revision 0.2  1995/10/24 16:32:20  glenn
 *	Revision 0.2
 *
 * Revision 0.1  1994/11/28  04:41:15  glenn
 * Initial release - supports DACs and counter/timers
 *
 */


#include <linux/config.h>

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif MODULE

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/config.h>
#include <linux/malloc.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>

#include "counter.h"
#include "dac.h"
#include "labpc.h"

static struct semaphore mutex = MUTEX;

static int active_dacs = 0;

static int write_dac( struct inode *inode, struct file *file,
		     const char *buffer, int count )
{
	int		i;
	Dac_t		*dac;
	short		*buf,	data;

	dprintk(( "LabPC+: write_dac()\n" ));

	/* Which DAC? */
	if ((dac = GETDAC( MINOR(inode->i_rdev) )) == NULL)
		return -(EINVAL);

	buf = (short *)kmalloc( count, GFP_KERNEL );
	memcpy_fromfs( buf, buffer, count );
	count /= sizeof( short );

	/* Lockout other processes from accessing the DAC */
	down( &(dac->sem) );

	/* Enable DAC updates on interrupts */
	if (dac->mode == DAC_TRACE_MODE || dac->mode == DAC_WAVEFORM_MODE
	    || dac->mode == DAC_TRACE_NOWAIT_MODE) {
		Counter_t *counter;

		cli();
		outb_p( 0, DAQ_TICLR );

		/* Set the counter for the DAC input */
		counter = GETCOUNTER( DAQ_COUNTERA2 );

		setbits( &(board->regs.cmd2), (LDAC0 << dac->num) );

		SetMode8253( counter, MODE_RATE_GENERATOR );
		Write8253( counter, dac->counter_val );

		if (dac->timebase != 2) {
			counter = GETCOUNTER( DAQ_COUNTERB0 );
			SetMode8253( counter, MODE_SQUARE_WAVE );
			Write8253( counter, dac->timebase );
			setbits( &(board->regs.cmd2), TBSEL );
		} else {
			clearbits( &(board->regs.cmd2), TBSEL );
		}

		/* Set the waveform buffer pointer */
		if (dac->buf)
			kfree( dac->buf );
		dac->buf = buf;
		dac->nbuf = count;
		dac->indx = 0;
		dac->trace = 0;
		dac->interrupt_flag = 1;
		active_dacs++;
		dac->nextval = dac->buf[0];
		if (dac->encoding == DAC_UNIPOLAR) {
			clearbits( &(board->regs.cmd2), TWOSDAC0 << dac->num );
		} else {
			setbits( &(board->regs.cmd2), TWOSDAC0 << dac->num );
		}

		/* Enable counter interrupts */
		setbits( &(board->regs.cmd3), CNTINTEN );

		if (dac->mode == DAC_TRACE_MODE) {
			interruptible_sleep_on( &(dac->wait) );
		}

		sti();
	} else {
		if (dac->interrupt_flag) {
			dac->interrupt_flag = 0;
			active_dacs--;

			/* Set the waveform buffer pointer */
			if (dac->buf)
				kfree( dac->buf );
			dac->buf = NULL;
			dac->nbuf = 0;
			dac->indx = 0;
			dac->trace = 0;
		}
		
		clearbits( &(board->regs.cmd2), (LDAC0 << dac->num) );
		clearbits( &(board->regs.cmd3), CNTINTEN );

		if (dac->encoding == DAC_UNIPOLAR) {
			clearbits( &(board->regs.cmd2), TWOSDAC0 << dac->num );
		} else {
			setbits( &(board->regs.cmd2), TWOSDAC0 << dac->num );
		}

		active_dacs++;

		for (i = 0; i < count; i++) {
			/* write the data to the DAC input buffer */
			data = dac->val = buf[i];
			outb_p( data & 0xff, dac->datalo );
			outb_p( (data >> 8) & 0xff, dac->datahi );
		}
		kfree( buf );

		outb_p( 0, DAQ_TICLR );
	}

	dprintregs();
	up( &(dac->sem) );

	return count * sizeof( short );
}

static int select_dac( struct inode *inode, struct file *file,
		      int sel_type, select_table *wait )
{
	Dac_t		*dac;
	int		retval;

	dprintk(( "LabPC+: select_counter()\n" ));
		
	/* Device always ready for read and write - wait on exceptions */
	if (sel_type != SEL_EX)
		return 0;		/* driver not ready */

	/* Which dac? */
	if ((dac = GETDAC( MINOR( inode->i_rdev ) )) == NULL)
		return -(EINVAL);

	/* Only wait if flagged to do so */
	if (wait == NULL)
		return (dac->interrupt_flag == 0);

	/* Register the wait queue for servicing by interrupts */
	retval = 0;
	cli();
	if (dac->interrupt_flag)
		select_wait( &(dac->wait), wait);
	else
		retval = 1;
	sti();
	
	return retval;			/* driver not ready */
}

static int ioctl_dac( struct inode *inode, struct file *file,
			 unsigned int cmd, unsigned long arg )
{
	Counter_t	*counter;
	Dac_t		*dac;
	int		stat;

	dprintk(( "LabPC+: ioctl_dac()\n" ));
		
	if ((stat = labpc_ioctl_check( cmd, arg )) != 0)
		return stat;
		
	/* Which counter? */
	if ((dac = GETDAC( MINOR(inode->i_rdev) )) == NULL)
		return -(EINVAL);

	/* Set the counter for the DAC input */
	counter = GETCOUNTER( DAQ_COUNTERA2 );

	/* Lockout other processes from accessing the DAC */
	down( &(dac->sem) );

	stat = 0;

	switch (cmd) {
	case DAC_SET_MODE:
		switch (arg) {
		case DAC_WRITE_MODE:
		case DAC_TRACE_MODE:
		case DAC_TRACE_NOWAIT_MODE:
		case DAC_WAVEFORM_MODE:
			dac->mode = arg;
			break;
		default:
			stat = -(EINVAL);
		}
		break;
	case DAC_GET_MODE:
		/* Get the DAC clock counter value */
		put_fs_long( dac->mode, (unsigned long *)arg );
		break;
	case DAC_SET_NUMTRACES:
		dac->ntraces = arg;
		break;
	case DAC_GET_NUMTRACES:
		/* Get the DAC clock counter value */
		put_fs_long( dac->ntraces, (unsigned long *)arg );
		break;
	case DAC_SET_CLOCK:
		dac->counter_val = arg;
		stat = Write8253( counter, arg );
		break;
	case DAC_GET_CLOCK:
		/* Get the DAC clock counter value */
		put_fs_long( dac->counter_val, (unsigned long *)arg );
		break;
	case DAC_START:
		dac->stop = 0;
		break;
	case DAC_STOP:
		dac->stop = 1;
		break;
	case DAC_SET_POLARITY:
		switch (arg) {
		case DAC_UNIPOLAR:
		case DAC_BIPOLAR:
			dac->encoding = arg;
			break;
		default:
			return -(EINVAL);
		}
		break;
	case DAC_GET_POLARITY:
		put_fs_long( dac->encoding, (unsigned long *)arg );
		break;
	case DAC_SET_TIMEBASE:
		if (arg == 2) {
			dac->timebase = arg;
			clearbits( &(board->regs.cmd2), TBSEL );
		} else if (arg > 2) {
			dac->timebase = arg;
			setbits( &(board->regs.cmd2), TBSEL );
		} else {
			stat = -(EINVAL);
		}
		break;
	case DAC_GET_TIMEBASE:
		put_fs_long( dac->timebase, (unsigned long *)arg );
		break;
	default:
		stat = -(EINVAL);
	}

	/* Unlock access to the dac registers */
	up( &(dac->sem) );

	return stat;
}

static int open_dac( struct inode *inode, struct file *file )
{
	Dac_t	*dac;
	int	minor;
	int	stat;

	dprintk(( "LabPC+: open_dac()\n" ));
		
	/* Which counter? */
	minor = MINOR(inode->i_rdev);
	if ((dac = GETDAC( minor )) == NULL)
		return -(EINVAL);

	if (dac->first_time) {
		dac->first_time	= 0;

		dac->open	= 0;
		dac->mode  	= DAC_WRITE_MODE;
		dac->sem   	= mutex;
		dac->wait  	= NULL;
		dac->buf   	= NULL;
		dac->nbuf  	= 0;
		dac->indx  	= 0;
		dac->counter_val= 0;
		dac->stop	= 0;
		dac->trace	= 0;
		dac->ntraces	= 1;
		dac->encoding	= DAC_BIPOLAR;
		dac->timebase	= 2;
		
		switch (minor) {
		case DAQ_DAC0:
			dac->num = 0;
			dac->datalo = DAQ_DAC0LO;
			dac->datahi = DAQ_DAC0HI;
			break;
		case DAQ_DAC1:
			dac->num = 1;
			dac->datalo = DAQ_DAC1LO;
			dac->datahi = DAQ_DAC1HI;
			break;
		default:
			printk( "LabPC+: Invalid DAC minor %d\n", minor );
			return -(EINVAL);
		}
	}
		
	/* This is an exclusive device */
	if (dac->open)
		return -(EBUSY);

	/* Grab the A2 counter for updating the DAC */
	if ((stat = Open8253( GETCOUNTER( DAQ_COUNTERA2 ) )) != 0)
		return stat;
	
	dac->open++;

	return 0;
}

static void release_dac( struct inode *inode, struct file *file )
{
	Dac_t 		*dac;
	int		minor;

	dprintk(( "LabPC+: release_dac()\n" ));
		
	minor = MINOR(inode->i_rdev);
	if ((dac = GETDAC( minor )) == NULL)
		return;

	Close8253( GETCOUNTER( DAQ_COUNTERA2 ) );

	dac->open--;
	labpc_close( inode, file );
}

void dac_interrupt( Dac_t *dac, int n )
{
	short	data;

	/* Handle interrupts for each of the DACs */
	while (n--) {
		if (dac->interrupt_flag && !dac->stop) {
			if (dac->indx >= dac->nbuf) {
				dac->indx = 0;
				dac->trace++;
				if ((dac->mode == DAC_TRACE_MODE  ||
				     dac->mode == DAC_TRACE_NOWAIT_MODE) &&
				    dac->trace >= dac->ntraces) {
					/* Wake up waiting tasks */
					dac->interrupt_flag = 0;
					if (!--active_dacs) {
						/* Cancel interrupts */
						clearbits( &(board->regs.cmd2),
							  (LDAC0 << dac->num) );
						clearbits( &(board->regs.cmd3),
							  CNTINTEN );
					}
					wake_up_interruptible( &(dac->wait) );
				}
			}
			/* Write next waveform value to the DAC */
			dac->val = dac->nextval;
			data = dac->nextval = dac->buf[dac->indx++];
			outb_p( data & 0xff, dac->datalo );
			outb_p( (data >> 8) & 0xff, dac->datahi );
		}
		dac++;
	}
	/* Clear the interrupt flag */
	outb_p( 0, DAQ_TICLR );
}

struct file_operations dac_fops = {
	NULL,				/* seek */
	NULL,				/* read */
	write_dac,
	NULL,				/* readdir */
	select_dac,
	ioctl_dac,
	NULL,				/* mmap */
	open_dac,
	release_dac,
};

