#include <stdio.h>
#include <stdlib.h>

#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>

#include "labpc.h"

#define MAXVALUES	10000


void Usage( void )
{
	fprintf( stderr, "Usage: writedac [-mode [wave,trace,write]] -numtraces n -wait -nowait -clock count -timebase microseconds -unipolar -bipolar] filename\n" );
}

int main( int argc, char **argv )
{
	unsigned short	value[MAXVALUES];
	float		v;
	char		*filename;
	int		d;
	int		mode = -1;
	int		clock = -1;
	int		nowait = -1;
	int		tbase = -1;
	int		polarity = DAC_BIPOLAR;
	int		ntrace = -1;
	int		fd;
	int		i;

	while (*++argv && **argv == '-') {
		if (!strcmp( *argv, "-mode" )) {
			if (!strcmp( *++argv, "wave" )) {
				mode = DAC_WAVEFORM_MODE;
			} else if (!strcmp( *argv, "trace" )) {
				mode = DAC_TRACE_MODE;
			} else if (!strcmp( *argv, "write" )) {
				mode = DAC_WRITE_MODE;
			} else {
				fprintf( stderr, "Invalid mode: %s\n", *argv );
				Usage();
				exit( 2 );
			}
		} else if (!strcmp( *argv, "-timebase" )) {
			if (sscanf( *++argv, "%g", &v ) != 1) {
				fprintf( stderr, "Invalid timebase: %s\n", *argv );
				Usage();
				exit( 2 );
			}
			tbase = v * 2.0 + 0.5;
		} else if (!strcmp( *argv, "-bipolar" )) {
			polarity = DAC_BIPOLAR;
		} else if (!strcmp( *argv, "-unipolar" )) {
			polarity = DAC_UNIPOLAR;
		} else if (!strcmp( *argv, "-nowait" )) {
			nowait = 1;
		} else if (!strcmp( *argv, "-wait" )) {
			nowait = 0;
		} else if (!strcmp( *argv, "-clock" )) {
			if (sscanf( *++argv, "%d", &clock ) != 1) {
				fprintf( stderr, "Invalid clock: %s\n", *argv );
				Usage();
				exit( 2 );
			}
		} else if (!strcmp( *argv, "-numtraces" )) {
			if (sscanf( *++argv, "%d", &ntrace ) != 1) {
				fprintf( stderr, "Invalid tracenum: %s\n", *argv );
				Usage();
				exit( 2 );
			}
		} else if (!strcmp( *argv, "-h" )) {
			Usage();
			exit( 1 );
		} else {
			fprintf( stderr, "Unknown option: %s\n", *argv );
			Usage();
			exit( 2 );
		}
	}

	if ((filename = *argv) == NULL) {
		Usage();
		exit( 1 );
	}

	if (nowait && mode == DAC_TRACE_MODE)
		mode = DAC_TRACE_NOWAIT_MODE;

	if ((fd = open( filename, O_RDWR )) < 0) {
		perror( filename );
		exit( 2 );
	}

	if (mode >= 0 && ioctl( fd, DAC_SET_MODE, mode ) < 0) {
		perror( filename );
		exit( 2 );
	}

	if (clock >= 0 && ioctl( fd, DAC_SET_CLOCK, clock ) < 0) {
		perror( filename );
		exit( 2 );
	}

	if (ntrace >= 0 && ioctl( fd, DAC_SET_NUMTRACES, ntrace ) < 0) {
		perror( filename );
		exit( 2 );
	}

	if (tbase >= 0 && ioctl( fd, DAC_SET_TIMEBASE, tbase ) < 0) {
		perror( filename );
		exit( 2 );
	}
		
	if (polarity >= 0 && ioctl( fd, DAC_SET_POLARITY, polarity ) < 0) {
		perror( filename );
		exit( 2 );
	}
		
	i = 0;
	while (scanf( "%g", &v ) == 1) {
		if (polarity == DAC_BIPOLAR) {
			d = (v / 5.0) * 2048;
			if (d > 2047)
				d = 2047;
			if (d < -2048)
				d = -2048;
		} else {
			d = (v / 10.0) * 4096;
			if (d > 4095)
				d = 4095;
			if (d < 0)
				d = 0;
		}
		value[i++] = d;
	}

	if (write( fd, value, i * sizeof( short ) ) < i * sizeof( short )) {
		perror( filename );
		exit( 2 );
	}

	close( fd );

	return 0;
}

	
	
	
	
