/*
 *	$Header: /home/glenn/src/labpc/RCS/labpcinit.c,v 0.2 1995/10/24 16:32:20 glenn Exp glenn $
 *
 *	$Log: labpcinit.c,v $
 *	Revision 0.2  1995/10/24 16:32:20  glenn
 *	Revision 0.2
 *
 * Revision 0.1  1994/11/28  04:41:15  glenn
 * Initial release - supports DACs and counter/timers
 *
 * Revision 1.2  1994/09/05  20:04:50  glenn
 * Fixed initialisation of Lab-PC+ card at start-up - may be more
 * reliable to detect now.
 *
 * Revision 1.1  1994/09/05  14:48:14  glenn
 * Initial revision
 *
 *
 */

#include <linux/config.h>

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#ifndef CONFIG_MODVERSIONS
char kernel_version[] = UTS_RELEASE;
#endif
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif MODULE

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/module.h>
#include <linux/malloc.h>
#include <linux/ioport.h>

#include <asm/io.h>
#include <asm/dma.h>
#include <asm/segment.h>
#include <asm/system.h>

#include <stdlib.h>

#include "adc.h"
#include "dac.h"
#include "counter.h"
#include "dio.h"
#include "labpc.h"

#define MAXINTERRUPTLOOP	10

#ifndef LABPC_MAJOR
#define LABPC_MAJOR		21
#endif

LabPC_t *board = NULL;

static int labpc_open( struct inode *inode, struct file *file )
{
	int	minor = MINOR( inode->i_rdev );
	int	stat = 0;

	dprintk(( "LabPC+ open called with minor %d\n", minor ));

	if (board == NULL)
		return -(ENODEV);

	if (DAQ_COUNTERA0 <= minor && minor <= DAQ_COUNTERB2) {
		file->f_op = &counter_fops;
	} else if (minor == DAQ_DAC0 || minor == DAQ_DAC1) {
		file->f_op = &dac_fops;
	} else if (minor == DAQ_ADC) {
		file->f_op = &adc_fops;
	} else if (DAQ_DPORTA <= minor && minor <= DAQ_DPORTC) {
		file->f_op = &dio_fops;
	} else if (minor != DAQ_LABPC) {
		return -(ENODEV);
	}
	stat = file->f_op->open( inode, file );
	MOD_INC_USE_COUNT;

	return stat;
}

void labpc_close( struct inode *inode, struct file *file )
{
	MOD_DEC_USE_COUNT;
}

int labpc_ioctl_check( unsigned int cmd, unsigned long arg )
{
	int	len,	err;

	if (cmd & IOC_INOUT) {
		/* Have to validate the address given by the process. */
		len = (cmd & IOCSIZE_MASK) >> IOCSIZE_SHIFT;
		
		if (cmd & IOC_IN) {
			if ((err = verify_area (VERIFY_READ, (void *) arg, len)) < 0)
				return err;
		}

		if (cmd & IOC_OUT) {
			if ((err = verify_area (VERIFY_WRITE, (void *) arg, len)) < 0)
				return err;
		}
	}

	return 0;
}

static int labpc_ioctl( struct inode *inode, struct file *file,
		       unsigned int cmd, unsigned long arg )
{
	int	stat = 0;
	
	if ((stat = labpc_ioctl_check( cmd, arg )) != 0)
		return stat;
		
	switch (cmd) {
	case LABPC_STATUS:
		/* Enable/disable ADC input channel scanning */
		dprintregs();

		break;
	default:
		stat = -(ENOSYS);
		break;
	}

	return stat;
}
			
static struct file_operations labpc_fops = {
        NULL,		/* seek */
	NULL,		/* read */
	NULL,		/* write */
	NULL,		/* readdir */
	NULL,		/* select */
	labpc_ioctl,	/* ioctl */
	NULL,		/* mmap */
        labpc_open,
        labpc_close	/* release */
};

static void daq_interrupt( int irq, struct pt_regs *registerp )
{
	int	i,	status;

	i = 0;
	do {
		status = board->stat = readport( &(board->regs.stat) );
		if (status & (DAVAIL | ADC_OVERRUN | ADC_OVERFLOW | DMATC)) {
			/* We have an interrupt for the ADC */
			adc_interrupt( board->adc, status );
		}
		if (status & CNTINT) {
			/* We have received an interrupt for the A2 counter. */
			dac_interrupt( board->dac, NUM_DACS );
		}
		if (++i > MAXINTERRUPTLOOP) {
			printk( "LabPC+: ISR looped %d times: status = 0x%x\n",
			       i, status );
			break;
		}
		status = board->stat = readport( &(board->regs.stat) );
	} while (status & (DAVAIL | ADC_OVERRUN | ADC_OVERFLOW
			    | DMATC | CNTINT));
}

int labpc_major	= LABPC_MAJOR;
int labpc_base 	= LABPC_BASE;
int labpc_irq	= LABPC_IRQ;
int labpc_dma	= LABPC_DMA;

int init_module( void )
{
	int 		i;
	CmdRegs_t 	*regs;

	if ((board = (LabPC_t *)kmalloc( sizeof( LabPC_t ), GFP_KERNEL )) == NULL)
		return -(ENOMEM);

	board->major   = labpc_major;
	board->base    = labpc_base;
	board->irq     = labpc_irq;
	board->dma     = labpc_dma;
	board->stat    = 0;
	board->open    = 0;
	board->present = 1;
	
	regs = &(board->regs);

	/* Check access to the io ports */
	if (check_region( board->base, LABPCIOSIZE )) {
		printk( "LabPC+: IO ports 0x%x to 0x%x already in use\n",
		        board->base, board->base + LABPCIOSIZE - 1 );
		printk( "LabPC+: Driver unloading\n" );

		return -(EIO);
	}

	/* Load the addresses of the command ports */
	regs->stat.address = DAQ_STAT;
	regs->cmd1.address = DAQ_CMD1;
	regs->cmd2.address = DAQ_CMD2;
	regs->cmd3.address = DAQ_CMD3;
	regs->cmd4.address = DAQ_CMD4;

	/*
	 * Initialise the Lab-PC+ Card (Manual 5-1). This is dodgy -
	 * we assume card is present before testing if this is
	 * so. However, there appears to be no reliable detection
	 * method, without first initialising the card.
	 */

	regs->stat.value = 0;
	writeport( &(regs->cmd1), 0x00 ); /* ADC gain & input channel	*/
	writeport( &(regs->cmd2), 0x00 ); /* DAC config & ADC trigger	*/
	writeport( &(regs->cmd3), 0x00 ); /* Disable interrupts & DMA	*/
	writeport( &(regs->cmd4), 0x00 ); /* Disable ADC interval scanning*/
	outb_p( 0x34, DAQ_CTAMODE );	/* Counter A0 output = HI	*/
	outb_p( 0x0a, DAQ_CTA0 );	/* Load init A0 counter value	*/
	outb_p( 0x00, DAQ_CTA0 );
	outb_p( 0x74, DAQ_CTAMODE );	/* Counter A1 output = HI	*/
	outb_p( 0xb4, DAQ_CTAMODE );	/* Counter A2 output = HI	*/
	outb_p( 0x34, DAQ_CTBMODE );	/* Counter B0 output = HI	*/
	outb_p( 0x74, DAQ_CTBMODE );	/* Counter B1 output = HI	*/
	outb_p( 0xb4, DAQ_CTBMODE );	/* Counter B2 output = HI	*/
	outb_p( 0x00, DAQ_DMATCCLR );	/* Clear DMA interrupt register	*/
	outb_p( 0x00, DAQ_TICLR );	/* Init timer int. clear reg.	*/
	outb_p( 0x00, DAQ_ADCCLR );	/* ADC Clear			*/
	inb_p( DAQ_FIFO );		/* Clear data out of ADC FIFO	*/
	inb_p( DAQ_FIFO );
	outb_p( 0x00, DAQ_DAC0LO );	/* Init value of DAC0		*/
	outb_p( 0x08, DAQ_DAC0HI );
	outb_p( 0x00, DAQ_DAC1LO );	/* Init value of DAC1		*/
	outb_p( 0x08, DAQ_DAC1HI );

	/* Check the Data available flag has been cleared */
	if ((i = readport( &(regs->stat) )) & DAVAIL) {
		printk( "Lab-PC+: Failed attempt to initialise board at base address: 0x%x\n",
		        board->base );
		return -(EIO);
	}
    
	board->counter = (Counter_t *)kmalloc( sizeof( Counter_t )
					      * NUM_COUNTERS, GFP_KERNEL );
	if (board->counter == NULL) {
		kfree( board );
		return -(ENOMEM);
	}

	for (i = 0; i < 3; i++) {
		Init8253( &(board->counter[i]),
			 DAQ_COUNTERA0 + i,
			 i,
			 DAQ_CTA0 + i,
			 DAQ_CTAMODE );
		Init8253( &(board->counter[3+i]),
			 DAQ_COUNTERB0 + i,
			 i,
			 DAQ_CTB0 + i,
			 DAQ_CTBMODE );
	}

	board->dac = (Dac_t *)kmalloc( sizeof( Dac_t ) * NUM_DACS, GFP_KERNEL );

	if (board->dac == NULL) {
		kfree( board->counter );
		kfree( board );
		return -(ENOMEM);
	}

	for (i = 0; i < NUM_DACS; i++) {
		board->dac[i].first_time = 1;
	}

	if ((board->adc = get_adc( DAQ_ADC )) == NULL) {
		kfree( board->counter );
		kfree( board->dac );
		kfree( board );
		return -(ENOMEM);
	}

	board->dio = (Dio_t *)kmalloc( sizeof( Dio_t ), GFP_KERNEL );
	if (board->dio == NULL) {
		kfree( board->counter );
		kfree( board->dac );
		kfree( board->adc );
		kfree( board );
		return -(ENOMEM);
	}
	InitDio( board->dio, DAQ_DPORTA, DAQ_DIGA,
		 DAQ_DIGB, DAQ_DIGC, DAQ_DIGMODE );

	/* Reserve the ports for our use */
	request_region( board->base, LABPCIOSIZE, "labpc" );
	
	if (register_chrdev( board->major, "labpc", &labpc_fops )) {
		printk( "LabPC+: Unable to get major %d for Lab-PC+ devices\n",
		       board->major );
		return -(EINVAL);
	} else {
		printk( "Lab-PC+ Card detected. Installed driver V0.1\n" );
	}

	if (request_dma( board->dma, "labpc" )) {
		printk( "LabPC+: can't request DMA channel %d\n", board->dma );
		return -(EIO);
	}
	if (request_irq( board->irq, daq_interrupt, 0, "labpc" )) {
		printk( "LabPC+: can't request IRQ %d\n", board->irq );
		return -(EIO);
	}

	return 0;
}

void cleanup_module(void)
{
	if (MOD_IN_USE) {
		printk("Lab-PC+: device busy, remove delayed\n");
	}

	free_adc( board->adc );
	dprintk(( "Lab-PC+: ADC released\n" ));
	
	free_irq( board->irq );
	dprintk(( "Lab-PC+: irq released\n" ));

	free_dma( board->dma );
	dprintk(( "Lab-PC+: dma released\n" ));
	
	release_region( board->base, LABPCIOSIZE );
	dprintk(( "Lab-PC+: ioports released\n" ));

	if (unregister_chrdev( board->major, "labpc" ) != 0) {
		printk( "Lab-PC+: device busy or other module error \n" );
	} else {
		printk( "Lab-PC+: succesfully removed \n" );
	} 
	dprintk(( "Lab-PC+: device major released\n" ));

	kfree( board->counter );
	dprintk(( "Lab-PC+: counter released\n" ));
	kfree( board->dac );
	dprintk(( "Lab-PC+: DAC released\n" ));
	kfree( board );
	dprintk(( "Lab-PC+: Board released\n" ));
}

